/*
 * Copyright 2005, 2006 University of Leiden.
 *
 * This file is part of MIA+EWS.
 *
 * MIA+EWS is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MIA+EWS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MIA; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/*******************************************************************************
* MPIA, NEVEC - MIDI project
*
* "@(#) $Id: source_table.c,v 1.4 2002/05/08 21:34:35 jdejong Exp $"
*
* who        when      what
* ---------  --------  ----------------------------------------------
* jdejong    2002-05-08 imported
*
*/
/* $Id: source_table.c,v 1.4 2002/05/08 21:34:35 jdejong Exp $ */
#define _POSIX_SOURCE_1
#include "vltPort.h"

/*----------------------------------------------------------------------- */
/* Source table for optical/IR interferometer                             */
/*----------------------------------------------------------------------- */
#include "source_table.h"

/* internal prototypes */
/* get data from table_def specific to source_table table */
void get_source_table_info (source_table_def *table);

/* file global variables for table type*/
static char *tabletype = "SOURCE";


/* source_table_row functions */
source_table_row* make_source_table_row(source_table_def* table)
/*----------------------------------------------------------------------- */
/*   Make an source_table_row structure                                   */
/*   Inputs:                                                              */
/*      table     pointer to source_table table structure                 */
/*   Returns:                                                             */
/*      An source_table_row structure pointer ; NULL indicates failure    */
/*      Data pointers set to generic structure values                     */
/*----------------------------------------------------------------------- */
{
  int nbands, i;
  source_table_row *out;

  /* create basic structure */
  out = (source_table_row*)malloc(sizeof(source_table_row));
  if (out==NULL) return out; /* allocation failure? */
  
  /* zero memory */
  memset (out, 0, sizeof(source_table_row));

  /* number of frequency bands */
  nbands = table->nbands;
  if (nbands<1) nbands = 1; /* at least one even if not used */
  out->iflux = (float*)malloc(nbands*sizeof(float));
  out->qflux = (float*)malloc(nbands*sizeof(float));
  out->uflux = (float*)malloc(nbands*sizeof(float));
  out->vflux = (float*)malloc(nbands*sizeof(float));
  out->freq = (double*)malloc(nbands*sizeof(double));
  for (i=0; i<nbands; i++) {
    out->freq[i] = 0.0;
    out->iflux[i] = 0.0;
    out->qflux[i] = 0.0;
    out->uflux[i] = 0.0;
    out->vflux[i] = 0.0;
  }
  
  /* Make generic row structure */
  out->raw = make_table_row_from_def(table->def);

  return out; /* done */
} /* end  make_source_table_row */

void kill_source_table_row (source_table_row *row)
/*----------------------------------------------------------------------- */
/*   Delete an source_table_row structure                                 */
/*   Inputs:                                                              */
/*      row    pointer to the source_table_row structure.                 */
/*----------------------------------------------------------------------- */
{
  if (row==NULL) return;         /* anybody home */
  kill_table_row(row->raw);      /* generic row structure */
  if (row->iflux!=NULL) free(row->iflux);  /* frequency info */
  if (row->qflux!=NULL) free(row->qflux);  /* frequency info */
  if (row->uflux!=NULL) free(row->uflux);  /* frequency info */
  if (row->vflux!=NULL) free(row->vflux);  /* frequency info */
  if (row->freq!=NULL) free(row->freq);    /* frequency info */
  free (row);                              /* basic structure */
} /* end  kill_source_table_row */


/* source_table functions */
source_table_def* 
make_source_table_def(fitsfile *fptr, int nbands, int *ierr)
/*----------------------------------------------------------------------- */
/*   Make a new source_table in an open FITS file                         */
/*   Inputs:                                                              */
/*      fptr      cfitsio pointer to open FITS file                       */
/*      nbands     number of values in freq, iflux...                     */
/*   Output:                                                              */
/*      ierr      Error code: 0 => ok, checked on input                   */
/*   Returns:                                                             */
/*      A source_table_def pointer; NULL indicates failure                */
/*----------------------------------------------------------------------- */
{
  source_table_def *out = NULL;
  int                    version = 1; /* table version number */
  int                    ikey, icol, i;
  long                   nx, ny, nregion;

  if (*ierr>0) return NULL;    /* existing error condition? */

  if (fptr==NULL)     return NULL; /* sanity check */

  /* create output object */
  out = (source_table_def*)malloc(sizeof(source_table_def));
  if (out==NULL)     {  /* check for allocation failure */
    *ierr = 10000;
    fprintf(stderr,"ERROR creating %s table_def \n", tabletype);
    return out;
  }
  memset (out, 0, sizeof(source_table_def)); /* zero fill */

  /* create table_def */
  out->def = make_table_def(SOURCE_TABLE, 0, 0);
  if (out->def==NULL) { /* check for failure */
    kill_source_table_def(out); /* clean up */
    *ierr = 10000;
    fprintf(stderr,"ERROR creating SOURCE_TABLE for %s table_def \n", 
	    tabletype);
    return NULL;
  }

  /* number of frequency bands */
  out->nbands = nbands;
  icol = table_find_col (out->def, "IFLUX")-1;
  if (icol>0) out->def->colcount[icol] = nbands;
  icol = table_find_col (out->def, "QFLUX")-1;
  if (icol>0) out->def->colcount[icol] = nbands;
  icol = table_find_col (out->def, "UFLUX")-1;
  if (icol>0) out->def->colcount[icol] = nbands;
  icol = table_find_col (out->def, "VFLUX")-1;
  if (icol>0) out->def->colcount[icol] = nbands;
  icol = table_find_col (out->def, "FREQ")-1;
  if (icol>0) out->def->colcount[icol] = nbands;

  /* add keyword values */
  ikey = table_find_key(out->def, "REVISION");
  if ((ikey>0) && (*ierr==0)) table_put_keyword (out->def, ikey, 
						(void*)&version, ierr);
  ikey = table_find_key(out->def, "NBANDS");
  if ((ikey>0) && (*ierr==0)) table_put_keyword (out->def, ikey, 
						(void*)&nbands, ierr);

  /* get info from table descriptor */
  get_source_table_info (out);

  /* open source table */
  make_table_from_def (fptr, out->def, ierr);

  return out; /* done */
} /* end make_source_table */

source_table_def* open_source_table(fitsfile *fptr, int *ierr)
/*----------------------------------------------------------------------- */
/*   Open an existing source_table in the active FITS file                */
/*   Inputs:                                                              */
/*      fptr      cfitsio pointer to open FITS file                       */
/*   Output:                                                              */
/*      ierr      Error code: 0 => ok, checked on input                   */
/*   Returns:                                                             */
/*      An source_table_def pointer; NULL indicates failure               */
/*----------------------------------------------------------------------- */
{
  source_table_def *out = NULL;

  if (*ierr>0) return NULL;    /* existing error condition? */
  if (fptr==NULL) return NULL; /* sanity check */

  /* create output object */
  out = (source_table_def*)malloc(sizeof(source_table_def));
  if (out==NULL)     {  /* check for allocation failure */
    *ierr = 10000;
    fprintf(stderr,"ERROR creating %s table_def \n", tabletype);
    return out;
  }
  memset (out, 0, sizeof(source_table_def)); /* zero fill */

  /* position FITS file at  table */
  if (fits_movnam_hdu(fptr, BINARY_TBL, tabletype, 0, ierr)) {
    fprintf(stderr,"ERROR %d finding %s table \n", *ierr, tabletype);
    kill_source_table_def(out); /* clean up */
    *ierr = 10000;
    return NULL;
  }

  /* get table descriptor */
  out->def = make_table_def_from_table (fptr);
  if (out->def==NULL) { /* error check */
    fprintf(stderr,"ERROR making table def for %s \n", tabletype);
    kill_source_table_def(out); /* clean up */
    *ierr = 10000;
    return NULL;
  }

  /* get info from table descriptor */
  get_source_table_info (out);

  return out; /* done */
} /* end open_source_table */

int read_source_table(fitsfile *fptr, source_table_def* table, 
			    source_table_row *row, long row_number,
			    int *ierr)
/*----------------------------------------------------------------------- */
/*   Read the next row from an source_table                               */
/*   Inputs:                                                              */
/*      fptr      cfitsio pointer to open FITS file                       */
/*      table     table definitions structure from open or create         */
/*      row_number which row (1-rel) in the table                         */   
/*                 -1 => next row                                         */  
/*   Input/output:                                                        */
/*      row       pointer to an extant row structure                      */
/*      ierr      Error code: 0 => ok, checked on input                   */
/*   Returns:                                                             */
/*      0 if successful, else failed                                      */
/*----------------------------------------------------------------------- */
{
  int i;
  table_row *trow;

  if (*ierr>0) return 1;     /* existing error condition? */
  if (fptr==NULL) return 1;  /* sanity check */
  if (table==NULL) return 1; /* sanity check */
  if (row==NULL) return 1;   /* sanity check */

  /* set row number */
  if (row_number>0) { /* explicit row number */
    row->raw->trow = row_number;
    table->row = row_number;
  } else { /* next row */
    row->raw->trow = ++table->row;
  }

  /* read row */
  table_read_row(fptr, table->def, row->raw, ierr);
  if (*ierr>0) { /* error check */
    fprintf(stderr,"ERROR  read_source_table: reading table %s \n", 
	    tabletype);
    return 1;
  }

  /* copy relevant data from generic table_row */
  trow = row->raw;
  row->raepp       = trow->data[table->raepp_col]->tdata.tdouble[0];
  row->decepp      = trow->data[table->decepp_col]->tdata.tdouble[0];
  row->rapp        = trow->data[table->rapp_col]->tdata.tdouble[0];
  row->decapp      = trow->data[table->decapp_col]->tdata.tdouble[0];
  row->sysvel      = trow->data[table->sysvel_col]->tdata.tdouble[0];
  row->pmra        = trow->data[table->pmra_col]->tdata.tdouble[0];
  row->pmdec       = trow->data[table->pmdec_col]->tdata.tdouble[0];
  row->parallax    = trow->data[table->parallax_col]->tdata.tfloat[0];
  row->diameter    = trow->data[table->diameter_col]->tdata.tfloat[0];
  row->diam_err    = trow->data[table->diam_err_col]->tdata.tfloat[0];
  row->source_id   = trow->data[table->source_id_col]->tdata.tshort[0];
  strncpy (row->source, trow->data[table->source_col]->tdata.tchar, 16);
  strncpy (row->calcode, trow->data[table->calcode_col]->tdata.tchar, 4);
  strncpy (row->equinox, trow->data[table->equinox_col]->tdata.tchar, 8);
  strncpy (row->veltyp, trow->data[table->veltyp_col]->tdata.tchar, 8);
  strncpy (row->veldef, trow->data[table->veldef_col]->tdata.tchar, 8);
  for (i=0; i<table->nbands; i++) {
    row->freq[i] = trow->data[table->freq_col]->tdata.tdouble[i];
    row->iflux[i] = trow->data[table->iflux_col]->tdata.tfloat[i];
    row->qflux[i] = trow->data[table->qflux_col]->tdata.tfloat[i];
    row->uflux[i] = trow->data[table->uflux_col]->tdata.tfloat[i];
    row->vflux[i] = trow->data[table->vflux_col]->tdata.tfloat[i];
  }

  return 0; /* OK - done */
} /* end read_source_table */

int write_source_table(fitsfile *fptr, source_table_def* table, 
			     source_table_row *row, long row_number, 
			     int *ierr)
/*----------------------------------------------------------------------- */
/*   Write  the next row to an source_table                               */
/*   Inputs:                                                              */
/*      fptr      cfitsio pointer to open FITS file                       */
/*      table     table definitions structure from open or create         */
/*      row       pointer to an extant row structure                      */
/*      row_number which row (1-rel) in the table                         */
/*                 -1 => next row                                         */  
/*   Input/output:                                                        */
/*      ierr      Error code: 0 => ok, checked on input                   */
/*   Returns:                                                             */
/*      0 if successful, else failed                                      */
/*----------------------------------------------------------------------- */
{
  table_row *trow;
  int        i;

  if (*ierr>0) return 1;     /* existing error condition? */
  if (fptr==NULL) return 1;  /* sanity check */
  if (table==NULL) return 1; /* sanity check */
  if (row==NULL) return 1;   /* sanity check */

  /* set row number */
  if (row_number>0) { /* explicit row number */
    row->raw->trow = row_number;
    table->row = row_number;
  } else { /* next row */
    row->raw->trow = ++table->row;
  }

  /* copy relevant data to generic table_row */
  trow = row->raw;
  trow->data[table->raepp_col]->tdata.tdouble[0]   = row->raepp;
  trow->data[table->decepp_col]->tdata.tdouble[0]  = row->decepp;
  trow->data[table->rapp_col]->tdata.tdouble[0]   = row->rapp;
  trow->data[table->decapp_col]->tdata.tdouble[0]  = row->decapp;
  trow->data[table->sysvel_col]->tdata.tdouble[0]  = row->sysvel;
  trow->data[table->pmra_col]->tdata.tdouble[0]    = row->pmra;
  trow->data[table->pmdec_col]->tdata.tdouble[0]   = row->pmdec;
  trow->data[table->parallax_col]->tdata.tfloat[0] = row->parallax;
  trow->data[table->diameter_col]->tdata.tfloat[0] = row->diameter;
  trow->data[table->diam_err_col]->tdata.tfloat[0] = row->diam_err;
  trow->data[table->source_id_col]->tdata.tshort[0]  = row->source_id;
  strncpy (trow->data[table->source_col]->tdata.tchar, row->source, 16);
  strncpy (trow->data[table->calcode_col]->tdata.tchar, row->calcode, 4);
  strncpy (trow->data[table->equinox_col]->tdata.tchar, row->equinox, 8);
  strncpy (trow->data[table->veltyp_col]->tdata.tchar, row->veltyp, 8);
  strncpy (trow->data[table->veldef_col]->tdata.tchar, row->veldef, 8);
  for (i=0; i<table->nbands; i++) {
    trow->data[table->freq_col]->tdata.tdouble[i] = row->freq[i];
    trow->data[table->iflux_col]->tdata.tfloat[i] = row->iflux[i];
    trow->data[table->qflux_col]->tdata.tfloat[i] = row->qflux[i];
    trow->data[table->uflux_col]->tdata.tfloat[i] = row->uflux[i];
    trow->data[table->vflux_col]->tdata.tfloat[i] = row->vflux[i];
  }

  /* write row */
  table_write_row(fptr, table->def, row->raw, ierr);
  if (*ierr>0) { /* error check */
    fprintf(stderr,"ERROR  write_source_table: writing table %s \n", 
	    tabletype);
    return 1;
  }

  return 0; /* OK - done */
} /* end write_source_table */

int close_source_table(fitsfile *fptr, source_table_def* table, 
			     int *ierr)
/*----------------------------------------------------------------------- */
/*   Close an open source_table, associated structures are destroyed      */
/*   Note: this closes the full FITS file                                 */
/*   Inputs:                                                              */
/*      fptr      cfitsio pointer to open FITS file                       */
/*   Returns:                                                             */
/*      0 if successful, else failed                                      */
/*----------------------------------------------------------------------- */
{
  if (*ierr>0) return 1;    /* existing error condition? */
  if (fptr==NULL) return 1; /* sanity check */

  /* do actual file close */
  fits_close_file (fptr, ierr);

  /* delete structures */
  kill_source_table_def(table);

  return 0; /* OK - done */
} /* end close_source_table */

void kill_source_table_def (source_table_def* table)
/*----------------------------------------------------------------------- */
/*   Delete an source_table structure                                     */
/*   Inputs:                                                              */
/*      table     table definitions structure from open or create         */
/*----------------------------------------------------------------------- */
{
  if (table==NULL) return; /* sanity check */

  /* table definition structure */
  kill_table_def(table->def);

  /* free basic structure */
  free (table);
} /* end kill_source_table_def */

/*************************private functions ******************************/

void get_source_table_info (source_table_def *table)
/*----------------------------------------------------------------------- */
/*   Get data from table_def specific to source_table table               */
/*   Uses detector member which must already be attached                  */
/*   Inputs:                                                              */
/*      table     table definitions structure                             */
/*----------------------------------------------------------------------- */
{
  int ikey, ierr=0;
  if (table==NULL) return;             /* sanity check */

  /* keyword values */
  ikey = table_find_key(table->def, "NBANDS");
  table->nbands = 0;
  if (ikey>0) 
    table_get_keyword(table->def, ikey, (void*)&table->nbands, &ierr);

  /* lookup column numbers (0-rel) */
  table->source_id_col = table_find_col (table->def, "SOURCE_ID")-1;
  table->source_col    = table_find_col (table->def, "SOURCE")-1;
  table->calcode_col   = table_find_col (table->def, "CALCODE")-1;
  table->iflux_col     = table_find_col (table->def, "IFLUX")-1;
  table->qflux_col     = table_find_col (table->def, "QFLUX")-1;
  table->uflux_col     = table_find_col (table->def, "UFLUX")-1;
  table->vflux_col     = table_find_col (table->def, "VFLUX")-1;
  table->freq_col      = table_find_col (table->def, "FREQ")-1;
  table->raepp_col     = table_find_col (table->def, "RAEPP")-1;
  table->decepp_col    = table_find_col (table->def, "DECEPP")-1;
  table->equinox_col   = table_find_col (table->def, "EQUINOX")-1;
  table->rapp_col      = table_find_col (table->def, "RAPP")-1;
  table->decapp_col    = table_find_col (table->def, "DECAPP")-1;
  table->sysvel_col    = table_find_col (table->def, "SYSVEL")-1;
  table->veltyp_col    = table_find_col (table->def, "VELTYP")-1;
  table->veldef_col    = table_find_col (table->def, "VELDEF")-1;
  table->pmra_col      = table_find_col (table->def, "PMRA")-1;
  table->pmdec_col     = table_find_col (table->def, "PMDEC")-1;
  table->parallax_col  = table_find_col (table->def, "PARALLAX")-1;
  table->diameter_col  = table_find_col (table->def, "DIAMETER")-1;
  table->diam_err_col  = table_find_col (table->def, "DIAM_ERR")-1;

} /* end  get_source_table_info */
/* */
